--[[ Copyright (c) 2009 Peter "Corsix" Cawley

Permission is hereby granted, free of charge, to any person obtaining a copy of
this software and associated documentation files (the "Software"), to deal in
the Software without restriction, including without limitation the rights to
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
of the Software, and to permit persons to whom the Software is furnished to do
so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE. ]]

--[[ Utility module for decoding Lua VM instructions ]]
--[[
  Main export is opcodes.decode(instruction)
  Also exports opcodes.names and opcodes.modes
--]]

require "util"

local enum = enum

module "opcodes"

names = enum {
-- Opcode names from lua's lopcodes.c
-- Behaviour of each opcode is listed in lopcodes.h, but is not reproduced here
-- as it is not required.
  "MOVE",
  "LOADK",
  "LOADBOOL",
  "LOADNIL",
  "GETUPVAL",
  "GETGLOBAL",
  "GETTABLE",
  "SETGLOBAL",
  "SETUPVAL",
  "SETTABLE",
  "NEWTABLE",
  "SELF",
  "ADD",
  "SUB",
  "MUL",
  "DIV",
  "MOD",
  "POW",
  "UNM",
  "NOT",
  "LEN",
  "CONCAT",
  "JMP",
  "EQ",
  "LT",
  "LE",
  "TEST",
  "TESTSET",
  "CALL",
  "TAILCALL",
  "RETURN",
  "FORLOOP",
  "FORPREP",
  "TFORLOOP",
  "SETLIST",
  "CLOSE",
  "CLOSURE",
  "VARARG",
}


local function opmode(t,a,b,c,mode)
  return {t = (t ~= 0), a = (a ~= 0), b = b, c = c, mode = mode}
end
local OpArgR = "R" -- register (iABC) or jump offset (iAsBx)
local OpArgK = "K" -- constant or register/constant
local OpArgU = "U" -- used (other)
local OpArgN = "N" -- not used
local iABC   = "ABC" -- One, two, or three fields
local iABx   = "ABx" -- B and C are combined to form extended B (Bx)
local iAsBx  = "AsBx" -- Bx is signed rather than unsigned (sBx)

modes = {
  [0] =
-- Opcode mode data from lua's lopcodes.c
--       T  A       B       C  mode         opcode
  opmode(0, 1, OpArgR, OpArgN, iABC)  -- /* OP_MOVE */
 ,opmode(0, 1, OpArgK, OpArgN, iABx)  -- /* OP_LOADK */
 ,opmode(0, 1, OpArgU, OpArgU, iABC)  -- /* OP_LOADBOOL */
 ,opmode(0, 1, OpArgR, OpArgN, iABC)  -- /* OP_LOADNIL */
 ,opmode(0, 1, OpArgU, OpArgN, iABC)  -- /* OP_GETUPVAL */
 ,opmode(0, 1, OpArgK, OpArgN, iABx)  -- /* OP_GETGLOBAL */
 ,opmode(0, 1, OpArgR, OpArgK, iABC)  -- /* OP_GETTABLE */
 ,opmode(0, 0, OpArgK, OpArgN, iABx)  -- /* OP_SETGLOBAL */
 ,opmode(0, 0, OpArgU, OpArgN, iABC)  -- /* OP_SETUPVAL */
 ,opmode(0, 0, OpArgK, OpArgK, iABC)  -- /* OP_SETTABLE */
 ,opmode(0, 1, OpArgU, OpArgU, iABC)  -- /* OP_NEWTABLE */
 ,opmode(0, 1, OpArgR, OpArgK, iABC)  -- /* OP_SELF */
 ,opmode(0, 1, OpArgK, OpArgK, iABC)  -- /* OP_ADD */
 ,opmode(0, 1, OpArgK, OpArgK, iABC)  -- /* OP_SUB */
 ,opmode(0, 1, OpArgK, OpArgK, iABC)  -- /* OP_MUL */
 ,opmode(0, 1, OpArgK, OpArgK, iABC)  -- /* OP_DIV */
 ,opmode(0, 1, OpArgK, OpArgK, iABC)  -- /* OP_MOD */
 ,opmode(0, 1, OpArgK, OpArgK, iABC)  -- /* OP_POW */
 ,opmode(0, 1, OpArgR, OpArgN, iABC)  -- /* OP_UNM */
 ,opmode(0, 1, OpArgR, OpArgN, iABC)  -- /* OP_NOT */
 ,opmode(0, 1, OpArgR, OpArgN, iABC)  -- /* OP_LEN */
 ,opmode(0, 1, OpArgR, OpArgR, iABC)  -- /* OP_CONCAT */
 ,opmode(0, 0, OpArgR, OpArgN, iAsBx) -- /* OP_JMP */
 ,opmode(1, 0, OpArgK, OpArgK, iABC)  -- /* OP_EQ */
 ,opmode(1, 0, OpArgK, OpArgK, iABC)  -- /* OP_LT */
 ,opmode(1, 0, OpArgK, OpArgK, iABC)  -- /* OP_LE */
 ,opmode(1, 1, OpArgR, OpArgU, iABC)  -- /* OP_TEST */
 ,opmode(1, 1, OpArgR, OpArgU, iABC)  -- /* OP_TESTSET */
 ,opmode(0, 1, OpArgU, OpArgU, iABC)  -- /* OP_CALL */
 ,opmode(0, 1, OpArgU, OpArgU, iABC)  -- /* OP_TAILCALL */
 ,opmode(0, 0, OpArgU, OpArgN, iABC)  -- /* OP_RETURN */
 ,opmode(0, 1, OpArgR, OpArgN, iAsBx) -- /* OP_FORLOOP */
 ,opmode(0, 1, OpArgR, OpArgN, iAsBx) -- /* OP_FORPREP */
 ,opmode(1, 0, OpArgN, OpArgU, iABC)  -- /* OP_TFORLOOP */
 ,opmode(0, 0, OpArgU, OpArgU, iABC)  -- /* OP_SETLIST */
 ,opmode(0, 0, OpArgN, OpArgN, iABC)  -- /* OP_CLOSE */
 ,opmode(0, 1, OpArgU, OpArgN, iABx)  -- /* OP_CLOSURE */
 ,opmode(0, 1, OpArgU, OpArgN, iABC)  -- /* OP_VARARG */
}

-- The following code assumes that instructions are laid out in the normal way:
--   32       23       14       6     0
-- ---+--------+--------+-------+-----+
--  0 |      B |      C |     A |  Op |
-- ---+--------+--------+-------+-----+
--  0 |           (s)Bx |     A |  Op |
-- ---+-----------------+-------+-----+

local max_sBx = (2 ^ 17) - 1

-- Converts a VM instruction encoded as an integer into a table containing the instruction
-- The returned table will have the following fields:
--  raw  - The instruction as an integer
--  op   - The opcode as an integer
--  name - The opcode as a string ("MOVE", "LOADK", etc.)
--  mode - The argument mode of the opcode ("ABC", "ABx", or "AsBx")
--  a    - The "A" field as an integer
--  b    - The "B", "Bx", or "sBx" field as a signed integer
--  c    - The "C" field as an integer, if there is one
function decode(instruction)
  local raw   =  instruction
  local op    =  instruction       % (2 ^ 6)
  instruction = (instruction - op) / (2 ^ 6)
  local a     =  instruction       % (2 ^ 8)
  instruction = (instruction - a ) / (2 ^ 8)
  
  local I = {op = op, a = a, name = names[op], mode = modes[op], raw = raw}
  local mode = modes[op].mode
  
  if mode == "ABC" then
    local c     =  instruction      % (2 ^ 9)
    instruction = (instruction - c) / (2 ^ 9)
    I.c = c
    I.b = instruction
  elseif mode == "ABx" then
    I.b = instruction
  elseif mode == "AsBx" then
    I.b = instruction - max_sBx
  else
    error "Unknown instruction mode"
  end
  
  return I
end
