--[[ Copyright (c) 2009 Peter "Corsix" Cawley

Permission is hereby granted, free of charge, to any person obtaining a copy of
this software and associated documentation files (the "Software"), to deal in
the Software without restriction, including without limitation the rights to
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
of the Software, and to permit persons to whom the Software is furnished to do
so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE. ]]

require "opcodes"

local error, opcodes_decode, setmetatable, ast, select, tostring, ipairs, assert, print
    = error, opcodes.decode, setmetatable, ast, select, tostring, ipairs, assert, print

module "undump"

function chunk(zio)
  local sig = zio:str(4)
  if sig ~= "\27Lua" then
    error "Invalid Lua signature"
  end

  local version          = zio:byte()
  local official         = zio:byte()
  local endian           = zio:byte()
  local size_int         = zio:byte()
  local size_size        = zio:byte()
  local size_instruction = zio:byte()
  local size_number      = zio:byte()
  local number_integral  = zio:byte()
  
  if version ~= 0x51 or official ~= 0 then
    error "Only Lua 5.1 official build is supported"
  end
  
  zio.endian = (endian == 1) and "little" or "big"
  zio:bind("int_t", "int", size_int)
  zio:bind("size_t", "uint", size_size)
  zio:bind("instruction", "uint", size_instruction)
  zio:bind("number", number_integral == 1 and "int" or "float", size_number)
  zio.string = function(self) return self:str(self:size_t()):sub(1, -2) end
  
  local v_minor = version % 16
  local v_major = (version - v_minor) / 16
  
  return
  {
    version = tostring(v_major) .. "." .. tostring(v_minor),
    official = (official == 0),
    endian = zio.endian,
    size_int = size_int,
    size_size_t = size_size,
    size_instruction = size_instruction,
    size_number = size_number,
    number_integral = number_integral,
    proto = prototype(zio),
  }
end

function prototype(zio)
  local P = {}
  P.source       = zio:string()
  P.start_line   = zio:int_t()
  P.end_line     = zio:int_t()
  P.num_upvalues = zio:byte()
  P.num_params   = zio:byte()
  P.is_vararg    = zio:byte()
  P.num_locals   = zio:byte()
  P.code = {}
  P.size_code = zio:int_t()
  for i = 1, P.size_code do
    P.code[i] = opcodes_decode(zio:instruction())
  end
  P.constants = {}
  P.size_constants = zio:int_t()
  for i = 1, P.size_constants do
    local t, v = zio:byte()
    if t == 0 then
      -- nil
    elseif t == 1 then
      v = zio:byte() ~= 0
    elseif t == 3 then
      v = zio:number()
    elseif t == 4 then
      v = zio:string()
    else
      error "bad constant"
    end
    P.constants[i - 1] = v
  end
  P.prototypes = {}
  P.size_prototypes = zio:int_t()
  for i = 1, P.size_prototypes do
    P.prototypes[i] = prototype(zio)
  end
  P.size_line_info = zio:int_t()
  P.line_info = {}
  for i = 1, P.size_line_info do
    P.line_info[i] = zio:int_t()
  end
  P.size_local_vars = zio:int_t()
  P.local_vars = {}
  for i = 1, P.size_local_vars do
    local T = {}
    T.name = zio:string()
    T.start_pc = zio:int_t()
    T.end_pc = zio:int_t()
    P.local_vars[i] = T
  end
  P.size_upvalue_names = zio:int_t()
  P.upvalue_names = {}
  for i = 1, P.size_upvalue_names do
    P.upvalue_names[i] = zio:string()
  end 
  
  return P
end
