--[[ Copyright (c) 2009 Peter "Corsix" Cawley

Permission is hereby granted, free of charge, to any person obtaining a copy of
this software and associated documentation files (the "Software"), to deal in
the Software without restriction, including without limitation the rights to
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
of the Software, and to permit persons to whom the Software is furnished to do
so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE. ]]

require "util"

local setmetatable, string_byte, string_sub, unpack, select, math_floor, table_reverse, error, ipairs
    = setmetatable, string.byte, string.sub, unpack, select, math.floor, table.reverse, error, ipairs

module "zio"

local zio_read_common = {}
local zio_read_string = setmetatable({}, {__index = zio_read_common})
local zio_read_string_mt = {
  __index = zio_read_string
}

function read_string(s)
  return setmetatable({buffer = s, index = 1, endian = "little"}, zio_read_string_mt)
end

function zio_read_string:eof()
  return self.index > #self.buffer
end

function zio_read_string:byte()
  if self:eof() then
    error "EOF reached"
  end
  local value = string_byte(self.buffer, self.index)
  self.index = self.index + 1
  return value
end

function zio_read_string:str(length)
  local new_index = self.index + length
  local end_index = new_index - 1
  if end_index > #self.buffer then
    error "EOF would be breached by read"
  end
  local value = string_sub(self.buffer, self.index, end_index)
  self.index = new_index
  return value
end

function zio_read_string:uint(length)
  if self.endian == "little" then
    local value, multiplier = 0, 1
    for i = 1, length do
      value = value + self:byte() * multiplier
      multiplier = multiplier * 256
    end
    return value
  else
    local value = 0
    for i = 1, length do
      value = value * 256 + self:byte()
    end
    return value
  end
end

function zio_read_common:int(length)
  local value = self:uint(length)
  if value >= 2 ^ (8 * length - 1) then
    value = value - (256 ^ length)
  end
  return value
end

function zio_read_common:bind(name, fn, ...)
  local args = {...}
  local n = select('#', ...)
  if n == 0 then
    self[name] = function()
      return self[fn](self)
    end
  elseif n == 1 then
    n = ...
    self[name] = function()
      return self[fn](self, n)
    end
  else
    self[name] = function()
      return self[fn](self, unpack(args, 1, n))
    end
  end
end

function zio_read_common:bitfield(...)
  local lengths = {...}
  local values = {}
  local bytes = {}
  local total_length = 0
  for _, length in ipairs(lengths) do
    total_length = total_length + length
  end
  table_reverse(lengths)
  for i = 1, math_floor((total_length + 7) / 8) do
    bytes[i] = self:byte()
  end
  if self.endian == "big" then
    table_reverse(bytes)
  end
  local byte_idx, bits_remaining, byte = 0, 0
  for i, length in ipairs(lengths) do
    local value, multiplier = 0, 1
    for _ = 1, length do
      if bits_remaining == 0 then
        byte_idx, bits_remaining = byte_idx + 1, 8
        byte = bytes[byte_idx]
      end
      local bit = byte % 2
      value = value + bit * multiplier
      byte = (byte - bit) / 2
      multiplier = multiplier * 2
      bits_remaining = bits_remaining - 1
    end
    values[i] = value
  end
  table_reverse(values)
  return unpack(values)
end

local float_ind  = 0 / 0      -- Generates -1.#IND on MSVC
local float_qnan = -float_ind -- Generates 1.#QNAN on MSVC
local float_inf  = 1 / 0      -- Generates 1.#INF on MSVC

function zio_read_common:float(length)
  local lengths
  if length == 4 then
    lengths = {1,  8, 23}
  elseif length == 8 then
    lengths = {1, 11, 52}
  else
    error "Unsupported floating point length"
  end
  
  local sign, exponent, fraction = self:bitfield(unpack(lengths))

  sign = (-1) ^ sign
  if exponent == 0 then
    if fraction == 0 then
      return sign == 1 and 0 or -0
    else
      exponent = exponent - 2 ^ (lengths[2] - 1) + 2 - lengths[3]
      if (0.4 == 0) and (exponent < 0) then
        return (sign * fraction) / 2 ^ (-exponent)
      else
        return sign * 2^exponent * fraction
      end
    end
  elseif exponent == (2 ^ lengths[2] - 1) then
    if fraction == 0 then
      return sign == 1 and float_inf or -float_inf
    else
      return float_qnan
    end
  else
    fraction = fraction + (2 ^ lengths[3])
    exponent = exponent - 2 ^ (lengths[2] - 1) + 1 - lengths[3]
    if (0.42 == 0) and (exponent < 0) then
      return (sign * fraction) / 2 ^ (-exponent)
    else
      return sign * 2^exponent * fraction
    end
  end
end
